#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
# Claire Delplancke (University of Bath)

from cil.optimisation.algorithms import Algorithm
from cil.optimisation.operators import BlockOperator
import numpy as np
import logging
from cil.optimisation.utilities import Sampler
from numbers import Number
import warnings
from cil.framework import BlockDataContainer

log = logging.getLogger(__name__)


class SPDHG(Algorithm):
    r'''Stochastic Primal Dual Hybrid Gradient (SPDHG) solves separable optimisation problems of the type: 
    
      .. math:: \min_{x} f(Kx) + g(x) = \min_{x} \sum f_i(K_i x) + g(x)

    where :math:`f_i` and the regulariser :math:`g` need to be proper, convex and lower semi-continuous.

    Parameters
    ----------
    f : BlockFunction
        Each must be a convex function with a "simple" proximal method of its conjugate
    g : Function
        A convex function with a "simple" proximal
    operator : BlockOperator
        BlockOperator must contain Linear Operators
    tau : positive float, optional
        Step size parameter for the primal problem. If `None` will be computed by algorithm, see note for details. 
    sigma : list of positive float, optional
        List of Step size parameters for dual problem.  If `None` will be computed by algorithm, see note for details. 
    initial : DataContainer, optional
        Initial point for the SPDHG algorithm. The default value is a zero DataContainer in the range of the `operator`.
    gamma : float, optional
            Parameter controlling the trade-off between the primal and dual step sizes
    sampler: `cil.optimisation.utilities.Sampler`, optional 
            A `Sampler` controllingthe selection of the next index for the SPDHG update. If `None`, a sampler will be created for uniform random sampling with replacement. See notes.  

    prob_weights: list of floats, optional,  
             Consider that the sampler is called a large number of times this argument holds the expected number of times each index would be called,  normalised to 1. Note that this should not be passed if the provided sampler has it as an attribute: if the sampler has a `prob_weight` attribute it will take precedence on this parameter. Should be a list of floats of length `num_indices` that sum to 1. If no sampler with `prob_weights` is passed, it defaults to `[1/len(operator)]*len(operator)`.


    Note  
    -----  
    The `sampler` can be an instance of the `cil.optimisation.utilities.Sampler` class or a custom class with the `__next__(self)` method implemented, which outputs an integer index from {1, ..., len(operator)}. 

    Note  
    -----  
    "Random sampling with replacement" will select the next index with equal probability from  `1 - len(operator)`.  


    Example
    -------
    >>> data = dataexample.SIMULATED_PARALLEL_BEAM_DATA.get()
    >>> data.reorder('astra')
    >>> data = data.get_slice(vertical='centre')
    >>> ig = ag.get_ImageGeometry()

    >>> data_partitioned = data.partition(num_batches=10, mode='staggered')
    >>> A_partitioned = ProjectionOperator(ig, data_partitioned.geometry, device = "cpu")


    >>> F = BlockFunction(*[L2NormSquared(b=data_partitioned[i])
                            for i in range(10)])

    >>> alpha = 0.025
    >>> G = alpha * TotalVariation()
    >>> spdhg = SPDHG(f=F, g=G, operator=A_partitioned, sampler=Sampler.sequential(len(A)),
                      initial=A.domain_geometry().allocate(1), update_objective_interval=10)
    >>> spdhg.run(100)


    Example
    -------
    Further examples of usage see the [CIL demos.](https://github.com/vais-ral/CIL-Demos/blob/master/Tomography/Simulated/Single%20Channel/PDHG_vs_SPDHG.py)

    Note
    -----
    When setting `sigma` and `tau`, there are 4 possible cases considered by setup function. In all cases the probabilities :math:`p_i` are set by a default or user defined sampler:

    - Case 1: If neither `sigma` or `tau` are provided then `sigma` is set using the formula:

    .. math:: \sigma_i= \frac{0.99}{\|K_i\|^2}

    and `tau` is set as per case 2

    - Case 2: If `sigma` is provided but not `tau` then `tau` is calculated using the formula 

    .. math:: \tau = 0.99\min_i( \frac{p_i}{ (\sigma_i  \|K_i\|^2) })

    - Case 3: If `tau` is provided but not `sigma` then `sigma` is calculated using the formula

    .. math:: \sigma_i= \frac{0.99 p_i}{\tau\|K_i\|^2}

    - Case 4: Both `sigma` and `tau` are provided.


    Note
    ----

    Convergence is guaranteed provided that [2, eq. (12)]:

    .. math:: \|\sigma[i]^{1/2}  K[i]  \tau^{1/2} \|^2  < p_i \text{ for all } i

    References
    ----------

    [1]"Stochastic primal-dual hybrid gradient algorithm with arbitrary 
    sampling and imaging applications",
    Chambolle, Antonin, Matthias J. Ehrhardt, Peter Richtárik, and Carola-Bibiane Schonlieb,
    SIAM Journal on Optimization 28, no. 4 (2018): 2783-2808.   https://doi.org/10.1137/17M1134834 

    [2]"Faster PET reconstruction with non-smooth priors by randomization and preconditioning",
    Matthias J Ehrhardt, Pawel Markiewicz and Carola-Bibiane Schönlieb,
    Physics in Medicine & Biology, Volume 64, Number 22, 2019. https://doi.org/10.1088/1361-6560/ab3d07
    '''

    def __init__(self, f=None, g=None, operator=None, tau=None, sigma=None,
                 initial=None, sampler=None, prob_weights=None,   **kwargs):

        update_objective_interval = kwargs.pop('update_objective_interval', 1)
        super(SPDHG, self).__init__(
            update_objective_interval=update_objective_interval)

        self.set_up(f=f, g=g, operator=operator, sigma=sigma, tau=tau,
                    initial=initial,  sampler=sampler, prob_weights=prob_weights,  **kwargs)

    def set_up(self, f, g, operator, sigma=None, tau=None,
               initial=None,   sampler=None, prob_weights=None, **deprecated_kwargs):
        '''set-up of the algorithm
        '''
        log.info("%s setting up", self.__class__.__name__)

        # algorithmic parameters
        self.f = f
        self.g = g
        self.operator = operator

        if not isinstance(operator, BlockOperator):
            raise TypeError("operator should be a BlockOperator")

        self._ndual_subsets = len(self.operator)
        
        self._prob_weights = getattr(sampler, 'prob_weights', prob_weights) 
        
        self._deprecated_set_prob(deprecated_kwargs, sampler) 
        
        if self._prob_weights is None: 
            self._prob_weights = [1/self._ndual_subsets]*self._ndual_subsets
        
        if  prob_weights is not None and self._prob_weights != prob_weights:
                    raise ValueError(' You passed a `prob_weights` argument and a sampler with a different attribute `prob_weights`, please remove the `prob_weights` argument.')

        if sampler is None:
            self._sampler = Sampler.random_with_replacement(
                len(operator), prob=self._prob_weights)
        else:
            self._sampler = sampler
        
        #Set the norms of the operators
        self._deprecated_set_norms(deprecated_kwargs) 
        self._norms = operator.get_norms_as_list()
        #Check for other kwargs
        if deprecated_kwargs:
            raise ValueError("Additional keyword arguments passed but not used: {}".format(deprecated_kwargs))

        self.set_step_sizes(sigma=sigma, tau=tau)

        # initialize primal variable
        if initial is None:
            self.x = self.operator.domain_geometry().allocate(0)
        else:
            self.x = initial.copy()

        self._x_tmp = self.operator.domain_geometry().allocate(0)

        # initialize dual variable to 0
        self._y_old = operator.range_geometry().allocate(0)
        if not isinstance(self._y_old, BlockDataContainer): #This can be removed once #1863 is fixed
            self._y_old =BlockDataContainer(self._y_old)

        # initialize variable z corresponding to back-projected dual variable
        self._z = operator.domain_geometry().allocate(0)
        self._zbar = operator.domain_geometry().allocate(0)
        # relaxation parameter
        self._theta = 1

        self.configured = True
        logging.info("{} configured".format(self.__class__.__name__, ))

    def _deprecated_set_prob(self, deprecated_kwargs, sampler):
        """
        Handle deprecated keyword arguments for backward compatibility.

        Parameters
        ----------
        deprecated_kwargs : dict
            Dictionary of keyword arguments.
        sampler : Sampler           
            Sampler class for selecting the next index for the SPDHG update.

        Notes
        -----
        This method is called by the set_up method.
        """
        
        prob = deprecated_kwargs.pop('prob', None)

        if prob is not None:
            if (self._prob_weights is None) and (sampler is None):
                warnings.warn('`prob` is being deprecated to be replaced with a sampler class and `prob_weights`. To randomly sample with replacement use "sampler=Sampler.randomWithReplacement(number_of_subsets,  prob=prob). To pass probabilities to the calculation for `sigma` and `tau` please use `prob_weights`. ', DeprecationWarning, stacklevel=2)
                self._prob_weights = prob
            else:

                raise ValueError(
                    '`prob` is being deprecated to be replaced with a sampler class and `prob_weights`. You passed  a `prob` argument, and either a `prob_weights` argument or a sampler. Please remove the `prob` argument.')



    def _deprecated_set_norms(self, deprecated_kwargs):
        """
        Handle deprecated keyword arguments for backward compatibility.

        Parameters
        ----------
        deprecated_kwargs : dict
            Dictionary of keyword arguments.

        Notes
        -----
        This method is called by the set_up method.
        """
        norms = deprecated_kwargs.pop('norms', None)
        
        if norms is not None:
            self.operator.set_norms(norms)
            warnings.warn(
                ' `norms` is being deprecated, use instead the `BlockOperator` function `set_norms`', DeprecationWarning, stacklevel=2)


        
    @property
    def sigma(self):
        return self._sigma

    @property
    def tau(self):
        return self._tau

    def set_step_sizes_from_ratio(self, gamma=1.0, rho=0.99):
        r""" Sets gamma, the step-size ratio for the SPDHG algorithm. Currently gamma takes a scalar value.

        The step sizes `sigma` and `tau` are set using the equations:

        .. math:: \sigma_i= \frac{\gamma\rho }{\|K_i\|^2}

        .. math::  \tau = \rho\min_i([ \frac{p_i }{\sigma_i  \|K_i\|^2})


        Parameters
        ----------
            gamma : Positive float
                parameter controlling the trade-off between the primal and dual step sizes
            rho : Positive float
                 parameter controlling the size of the product :math:`\sigma\tau`



        """
        if isinstance(gamma, Number):
            if gamma <= 0:
                raise ValueError(
                    "The step-sizes of SPDHG are positive, gamma should also be positive")

        else:
            raise ValueError(
                "We currently only support scalar values of gamma")
        if isinstance(rho, Number):
            if rho <= 0:
                raise ValueError(
                    "The step-sizes of SPDHG are positive, rho should also be positive")

        else:
            raise ValueError(
                "We currently only support scalar values of gamma")

        self._sigma = [gamma * rho / ni for ni in self._norms]
        values = [rho*pi / (si * ni**2) for pi, ni,
                  si in zip(self._prob_weights, self._norms, self._sigma)]
        self._tau = min([value for value in values if value > 1e-8])

    def set_step_sizes(self, sigma=None, tau=None):
        r""" Sets sigma and tau step-sizes for the SPDHG algorithm after the initial set-up. The step sizes can be either scalar or array-objects.

        When setting `sigma` and `tau`, there are 4 possible cases considered by setup function: 

        - Case 1: If neither `sigma` or `tau` are provided then `sigma` is set using the formula:

        .. math:: \sigma_i= \frac{0.99}{\|K_i\|^2}

        and `tau` is set as per case 2

        - Case 2: If `sigma` is provided but not `tau` then `tau` is calculated using the formula 

        .. math:: \tau = 0.99\min_i( \frac{p_i}{ (\sigma_i  \|K_i\|^2) })

        - Case 3: If `tau` is provided but not `sigma` then `sigma` is calculated using the formula

        .. math:: \sigma_i= \frac{0.99 p_i}{\tau\|K_i\|^2}

        - Case 4: Both `sigma` and `tau` are provided.


        Parameters
        ----------
            sigma : list of positive float, optional, default= see docstring
                List of Step size parameters for dual problem
            tau : positive float, optional, default= see docstring
                Step size parameter for primal problem

        """
        gamma = 1.
        rho = .99
        if sigma is not None:
            if len(sigma) == self._ndual_subsets:
                if all(isinstance(x, Number) and x > 0 for x in sigma):
                    pass
                else:
                    raise ValueError(
                        "Sigma expected to be a positive number.")

            else:
                raise ValueError(
                    "Please pass a list of floats to sigma with the same number of entries as number of operators")
            self._sigma = sigma

        elif tau is None:
            self._sigma = [gamma * rho / ni for ni in self._norms]
        else:
            self._sigma = [
                rho*pi / (tau*ni**2) for ni, pi in zip(self._norms, self._prob_weights)]

        if tau is None:
            values = [rho*pi / (si * ni**2) for pi, ni,
                      si in zip(self._prob_weights, self._norms, self._sigma)]
            self._tau = min([value for value in values if value > 1e-8])

        else:
            if not ( isinstance(tau, Number) and tau > 0):
                raise ValueError(
                    "The step-sizes of SPDHG must be positive, passed tau = {}".format(tau))

            self._tau = tau

    def check_convergence(self):
        """  Checks whether convergence criterion for SPDHG is satisfied with the current scalar values of tau and sigma

        Returns
        -------
        Boolean
            True if convergence criterion is satisfied. False if not satisfied or convergence is unknown. 

        Note
        -----
        Convergence criterion currently can only be checked for scalar values of tau.

        Note
        ----
        This checks the convergence criterion. Numerical errors may mean some sigma and tau values that satisfy the convergence criterion may not converge. 
        Alternatively, step sizes outside the convergence criterion may still allow (fast) convergence. 
        """
        for i in range(self._ndual_subsets):
            if isinstance(self._tau, Number) and isinstance(self._sigma[i], Number):
                if self._sigma[i] * self._tau * self._norms[i]**2 > self._prob_weights[i]:
                    return False
                return True
            else:
                raise ValueError('Convergence criterion currently can only be checked for scalar values of tau and sigma[i].')

    def update(self):
        """  Runs one iteration of SPDHG 

        """
        # Gradient descent for the primal variable
        # x_tmp = x - tau * zbar
        self._zbar.sapyb(self._tau,  self.x, -1., out=self._x_tmp )
        self._x_tmp*=-1

        self.g.proximal(self._x_tmp, self._tau, out=self.x)

        # Choose subset
        i = next(self._sampler)

        # Gradient ascent for the dual variable
        # y_k = y_old[i] + sigma[i] * K[i] x
        try:
            y_k = self.operator[i].direct(self.x)
        except IndexError:
            raise IndexError(
                'The sampler has outputted an index larger than the number of operators to sample from. Please ensure your sampler samples from {0,1,...,len(operator)-1} only.')

        y_k.sapyb(self._sigma[i], self._y_old[i], 1., out=y_k)

        y_k = self.f[i].proximal_conjugate(y_k, self._sigma[i])

        # Back-project
        # x_tmp = K[i]^*(y_k - y_old[i])
        y_k.subtract(self._y_old[i], out=self._y_old[i])

        self.operator[i].adjoint(self._y_old[i], out=self._x_tmp)
        # Update backprojected dual variable and extrapolate
        # zbar = z + (1 + theta/p[i]) x_tmp

        # z = z + x_tmp
        self._z.add(self._x_tmp, out=self._z)
        # zbar = z + (theta/p[i]) * x_tmp

        self._z.sapyb(1., self._x_tmp, self._theta /
                      self._prob_weights[i], out=self._zbar)

        # save previous iteration
        self._save_previous_iteration(i, y_k)

    def update_objective(self):
        # p1 = self.f(self.operator.direct(self.x)) + self.g(self.x)
        p1 = 0.
        for i, op in enumerate(self.operator.operators):
            p1 += self.f[i](op.direct(self.x))
        p1 += self.g(self.x)

        d1 = - self.f.convex_conjugate(self._y_old)
        tmp = self.operator.adjoint(self._y_old)
        tmp *= -1
        d1 -= self.g.convex_conjugate(tmp)

        self.loss.append([p1, d1, p1-d1])

    @property
    def objective(self):
        '''The saved primal objectives. 

        Returns
        -------
        list
            The saved primal objectives from `update_objective`. The number of saved values depends on the `update_objective_interval` kwarg. 
        '''
        return [x[0] for x in self.loss]

    @property
    def dual_objective(self):
        '''The saved dual objectives. 

        Returns
        -------
        list
            The saved dual objectives from `update_objective`. The number of saved values depends on the `update_objective_interval` kwarg. 
        '''
        return [x[1] for x in self.loss]

    @property
    def primal_dual_gap(self):
        '''The saved primal-dual gap. 

        Returns
        -------
        list
            The saved primal dual gap from `update_objective`. The number of saved values depends on the `update_objective_interval` kwarg. 
        '''
        return [x[2] for x in self.loss]

    def _save_previous_iteration(self, index, y_current):
        ''' Internal function used to save the previous iteration 
        '''
        self._y_old[index].fill(y_current)



#  Copyright 2020 United Kingdom Research and Innovation
#  Copyright 2020 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt


from cil.framework import DataProcessor, ImageData
from cil.plugins.astra.utilities import convert_geometry_to_astra_vec_3D
import astra
import numpy as np


class FDK_Flexible(DataProcessor):

    """
    FDK_Flexible Filtered Back Projection performs an FDK reconstruction for 2D and 3D cone-beam geometries.
    It is able to back-project circular trajectories with 2 PI angular range and equally spaced angular steps.

    This uses the ram-lak filter
    This is a GPU version only

    Parameters
    ----------
    volume_geometry : ImageGeometry
        A description of the area/volume to reconstruct

    sinogram_geometry : AcquisitionGeometry
        A description of the acquisition data

    Example
    -------
    >>> from cil.plugins.astra import FDK_Flexible
    >>> fbp = FDK_Flexible(image_geometry, data.geometry)
    >>> fbp.set_input(data)
    >>> reconstruction = fbp.get_ouput()

    """

    def __init__(self, volume_geometry,
                       sinogram_geometry):

        vol_geom_astra, proj_geom_astra = convert_geometry_to_astra_vec_3D(volume_geometry, sinogram_geometry)


        super(FDK_Flexible, self).__init__( volume_geometry = volume_geometry,
                                        sinogram_geometry = sinogram_geometry,
                                        vol_geom_astra = vol_geom_astra,
                                        proj_geom_astra = proj_geom_astra)



    def check_input(self, dataset):

        if self.sinogram_geometry.channels != 1:
            raise ValueError("Expected input data to be single channel, got {0}"\
                 .format(self.sinogram_geometry.channels))

        if self.sinogram_geometry.geom_type != 'cone':
            raise ValueError("Expected input data to be cone beam geometry , got {0}"\
                 .format(self.sinogram_geometry.geom_type))

        return True
    
    def _set_up(self):
        """
        Configure processor attributes that require the data to setup
        Must set _shape_out
        """
        self._shape_out = self.volume_geometry.shape
    
    def process(self, out=None):

        # Get DATA
        DATA = self.get_input()

        pad = False
        if len(DATA.shape) == 2:
            #for 2D cases
            pad = True
            data_temp = np.expand_dims(DATA.as_array(),axis=0)
        else:
            data_temp = DATA.as_array()


        rec_id = astra.data3d.create('-vol', self.vol_geom_astra)
        sinogram_id = astra.data3d.create('-sino', self.proj_geom_astra, data_temp)
        cfg = astra.astra_dict('FDK_CUDA')
        cfg['ReconstructionDataId'] = rec_id
        cfg['ProjectionDataId'] = sinogram_id
        alg_id = astra.algorithm.create(cfg)

        astra.algorithm.run(alg_id)
        arr_out = astra.data3d.get(rec_id)

        astra.data3d.delete(rec_id)
        astra.data3d.delete(sinogram_id)
        astra.algorithm.delete(alg_id)

        if pad == True:
            arr_out = np.squeeze(arr_out, axis=0)

        if out is None:
            out = ImageData(arr_out, deep_copy=False, geometry=self.volume_geometry.copy(), suppress_warning=True)
            return out
        else:
            out.fill(arr_out)



#  Copyright 2018 United Kingdom Research and Innovation
#  Copyright 2018 The University of Manchester
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
# Authors:
# CIL Developers, listed at: https://github.com/TomographicImaging/CIL/blob/master/NOTICE.txt
import copy
import warnings
from numbers import Number

import numpy

from .image_data import ImageData
from .labels import ImageDimension, FillType


class ImageGeometry:
    @property
    def CHANNEL(self):
        warnings.warn("use ImageDimension.CHANNEL instead", DeprecationWarning, stacklevel=2)
        return ImageDimension.CHANNEL

    @property
    def HORIZONTAL_X(self):
        warnings.warn("use ImageDimension.HORIZONTAL_X instead", DeprecationWarning, stacklevel=2)
        return ImageDimension.HORIZONTAL_X

    @property
    def HORIZONTAL_Y(self):
        warnings.warn("use ImageDimension.HORIZONTAL_Y instead", DeprecationWarning, stacklevel=2)
        return ImageDimension.HORIZONTAL_Y

    @property
    def RANDOM(self):
        warnings.warn("use FillType.RANDOM instead", DeprecationWarning, stacklevel=2)
        return FillType.RANDOM
    @property
    def RANDOM_INT(self):
        warnings.warn("use FillType.RANDOM_INT instead", DeprecationWarning, stacklevel=2)
        return FillType.RANDOM_INT

    @property
    def VERTICAL(self):
        warnings.warn("use ImageDimension.VERTICAL instead", DeprecationWarning, stacklevel=2)
        return ImageDimension.VERTICAL

    @property
    def shape(self):
        shape_dict = {ImageDimension.CHANNEL: self.channels,
                      ImageDimension.VERTICAL: self.voxel_num_z,
                      ImageDimension.HORIZONTAL_Y: self.voxel_num_y,
                      ImageDimension.HORIZONTAL_X: self.voxel_num_x}
        return tuple(shape_dict[label] for label in self.dimension_labels)

    @shape.setter
    def shape(self, val):
        print("Deprecated - shape will be set automatically")

    @property
    def spacing(self):
        spacing_dict = {ImageDimension.CHANNEL: self.channel_spacing,
                        ImageDimension.VERTICAL: self.voxel_size_z,
                        ImageDimension.HORIZONTAL_Y: self.voxel_size_y,
                        ImageDimension.HORIZONTAL_X: self.voxel_size_x}
        return tuple(spacing_dict[label] for label in self.dimension_labels)

    @property
    def length(self):
        return len(self.dimension_labels)

    @property
    def ndim(self):
        return len(self.dimension_labels)

    @property
    def dimension_labels(self):

        labels_default = ImageDimension.get_order_for_engine("cil")

        shape_default = [   self.channels,
                            self.voxel_num_z,
                            self.voxel_num_y,
                            self.voxel_num_x]

        try:
            labels = self._dimension_labels
        except AttributeError:
            labels = labels_default
        labels = list(labels)

        for i, x in enumerate(shape_default):
            if x == 0 or x==1:
                try:
                    labels.remove(labels_default[i])
                except ValueError:
                    pass #if not in custom list carry on
        return tuple(labels)

    @dimension_labels.setter
    def dimension_labels(self, val):
        self.set_labels(val)

    def set_labels(self, labels):
        if labels is not None:
            self._dimension_labels = tuple(map(ImageDimension, labels))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        if self.voxel_num_x == other.voxel_num_x \
            and self.voxel_num_y == other.voxel_num_y \
            and self.voxel_num_z == other.voxel_num_z \
            and self.voxel_size_x == other.voxel_size_x \
            and self.voxel_size_y == other.voxel_size_y \
            and self.voxel_size_z == other.voxel_size_z \
            and self.center_x == other.center_x \
            and self.center_y == other.center_y \
            and self.center_z == other.center_z \
            and self.channels == other.channels \
            and self.channel_spacing == other.channel_spacing \
            and self.dimension_labels == other.dimension_labels:

            return True

        return False

    @property
    def dtype(self):
        return self._dtype

    @dtype.setter
    def dtype(self, val):
        self._dtype = val

    def __init__(self,
                 voxel_num_x=0,
                 voxel_num_y=0,
                 voxel_num_z=0,
                 voxel_size_x=1,
                 voxel_size_y=1,
                 voxel_size_z=1,
                 center_x=0,
                 center_y=0,
                 center_z=0,
                 channels=1,
                 **kwargs):

        self.voxel_num_x = int(voxel_num_x)
        self.voxel_num_y = int(voxel_num_y)
        self.voxel_num_z = int(voxel_num_z)
        self.voxel_size_x = float(voxel_size_x)
        self.voxel_size_y = float(voxel_size_y)
        self.voxel_size_z = float(voxel_size_z)
        self.center_x = center_x
        self.center_y = center_y
        self.center_z = center_z
        self.channels = channels
        self.channel_labels = None
        self.channel_spacing = 1.0
        self.dimension_labels = kwargs.get('dimension_labels', None)
        self.dtype = kwargs.get('dtype', numpy.float32)


    def get_slice(self,channel=None, vertical=None, horizontal_x=None, horizontal_y=None):
        '''
        Returns a new ImageGeometry of a single slice of in the requested direction.
        '''
        geometry_new = self.copy()
        if channel is not None:
            geometry_new.channels = 1

            try:
                geometry_new.channel_labels = [self.channel_labels[channel]]
            except:
                geometry_new.channel_labels = None

        if vertical is not None:
            geometry_new.voxel_num_z = 0

        if horizontal_y is not None:
            geometry_new.voxel_num_y = 0

        if horizontal_x is not None:
            geometry_new.voxel_num_x = 0

        return geometry_new

    def get_order_by_label(self, dimension_labels, default_dimension_labels):
        order = []
        for i, el in enumerate(default_dimension_labels):
            for j, ek in enumerate(dimension_labels):
                if el == ek:
                    order.append(j)
                    break
        return order

    def get_min_x(self):
        return self.center_x - 0.5*self.voxel_num_x*self.voxel_size_x

    def get_max_x(self):
        return self.center_x + 0.5*self.voxel_num_x*self.voxel_size_x

    def get_min_y(self):
        return self.center_y - 0.5*self.voxel_num_y*self.voxel_size_y

    def get_max_y(self):
        return self.center_y + 0.5*self.voxel_num_y*self.voxel_size_y

    def get_min_z(self):
        if not self.voxel_num_z == 0:
            return self.center_z - 0.5*self.voxel_num_z*self.voxel_size_z
        else:
            return 0

    def get_max_z(self):
        if not self.voxel_num_z == 0:
            return self.center_z + 0.5*self.voxel_num_z*self.voxel_size_z
        else:
            return 0

    def clone(self):
        '''returns a copy of the ImageGeometry'''
        return copy.deepcopy(self)

    def copy(self):
        '''alias of clone'''
        return self.clone()

    def __str__ (self):
        repres = ""
        repres += "Number of channels: {0}\n".format(self.channels)
        repres += "channel_spacing: {0}\n".format(self.channel_spacing)

        if self.voxel_num_z > 0:
            repres += "voxel_num : x{0},y{1},z{2}\n".format(self.voxel_num_x, self.voxel_num_y, self.voxel_num_z)
            repres += "voxel_size : x{0},y{1},z{2}\n".format(self.voxel_size_x, self.voxel_size_y, self.voxel_size_z)
            repres += "center : x{0},y{1},z{2}\n".format(self.center_x, self.center_y, self.center_z)
        else:
            repres += "voxel_num : x{0},y{1}\n".format(self.voxel_num_x, self.voxel_num_y)
            repres += "voxel_size : x{0},y{1}\n".format(self.voxel_size_x, self.voxel_size_y)
            repres += "center : x{0},y{1}\n".format(self.center_x, self.center_y)

        return repres
    def allocate(self, value=0, **kwargs):
        '''allocates an ImageData according to the size expressed in the instance

        :param value: accepts numbers to allocate an uniform array, or a string as 'random' or 'random_int' to create a random array or None.
        :type value: number or string, default None allocates empty memory block, default 0
        :param dtype: numerical type to allocate
        :type dtype: numpy type, default numpy.float32
        '''

        dtype = kwargs.get('dtype', self.dtype)

        if kwargs.get('dimension_labels', None) is not None:
            raise ValueError("Deprecated: 'dimension_labels' cannot be set with 'allocate()'. Use 'geometry.set_labels()' to modify the geometry before using allocate.")

        out = ImageData(geometry=self.copy(),
                            dtype=dtype,
                            suppress_warning=True)

        if isinstance(value, Number):
            # it's created empty, so we make it 0
            out.array.fill(value)
        elif value in FillType:
            if value == FillType.RANDOM:
                seed = kwargs.get('seed', None)
                if seed is not None:
                    numpy.random.seed(seed)
                if numpy.iscomplexobj(out.array):
                    r = numpy.random.random_sample(self.shape) + 1j * numpy.random.random_sample(self.shape)
                    out.fill(r)
                else:
                    out.fill(numpy.random.random_sample(self.shape))

            elif value == FillType.RANDOM_INT:
                seed = kwargs.get('seed', None)
                if seed is not None:
                    numpy.random.seed(seed)
                max_value = kwargs.get('max_value', 100)
                if numpy.iscomplexobj(out.array):
                    out.fill(numpy.random.randint(max_value,size=self.shape, dtype=numpy.int32) + 1.j*numpy.random.randint(max_value,size=self.shape, dtype=numpy.int32))
                else:
                    out.fill(numpy.random.randint(max_value,size=self.shape, dtype=numpy.int32))
        elif value is None:
            pass
        else:
            raise ValueError(f'Value {value} unknown')
        return out




class InPlaceError(Exception):
    "Raised when the object should not be used in-place"
    def __init__(self,  message="Object should not be used in-place"):
        self.message = message
        super().__init__(self.message)



import unittest
import numpy as np
from cil.utilities.display import show1D
from cil.utilities import dataexample
from unittest.mock import patch
from utils import initialise_tests
initialise_tests()

import matplotlib

class TestShow1D(unittest.TestCase):

    def setUp(self):
        # Set up example data for testing
        self.data = dataexample.SIMULATED_SPHERE_VOLUME.get()
        self.data2 = self.data * 0.8
        self.data3 = np.random.rand(10, 10, 10)


    @patch('matplotlib.pyplot.show')
    def test_show1D_figure_singledata(self, mock_show):
        # Test all components of the plot

        try:
            fig = show1D(self.data, slice_list=[(0, 64),(1,64)])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with all plot components: {e}")

        # Check the data
        line = fig.figure.axes[0].lines[0]
        x_data = line.get_xdata()
        y_data = line.get_ydata()
        self.assertEqual(len(x_data), 128)
        self.assertEqual(len(y_data), 128)
        np.testing.assert_array_equal(y_data, self.data.array[64, 64, :])

        # Check the plot components
        self.assertEqual(len(fig.figure.axes), 1)
        self.assertEqual(fig.figure.axes[0].get_title(), 'Slice at vertical:64, horizontal_y:64')
        self.assertEqual(fig.figure.axes[0].get_xlabel(), 'horizontal_x index')
        self.assertEqual(fig.figure.axes[0].get_ylabel(), 'Value')
        self.assertEqual(fig.figure.axes[0].get_xlim(), (0, 127))


    @patch('matplotlib.pyplot.show')
    def test_show1D_figure_multidata(self, mock_show):
        # Test all components of the plot

        try:
            fig = show1D([self.data, self.data2], slice_list=[(0, 64),(1,64)])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with all plot components: {e}")

        # Check the data
        line = fig.figure.axes[0].lines[0]
        x_data = line.get_xdata()
        y_data = line.get_ydata()
        self.assertEqual(len(x_data), 128)
        self.assertEqual(len(y_data), 128)
        np.testing.assert_array_equal(y_data, self.data.array[64, 64, :])

        line = fig.figure.axes[0].lines[1]
        x_data = line.get_xdata()
        y_data = line.get_ydata()
        self.assertEqual(len(x_data), 128)
        self.assertEqual(len(y_data), 128)
        np.testing.assert_array_equal(y_data, self.data2.array[64, 64, :])

        # Check the plot components
        self.assertEqual(len(fig.figure.axes), 1)
        self.assertEqual(fig.figure.axes[0].get_title(), 'Slice at vertical:64, horizontal_y:64')
        self.assertEqual(fig.figure.axes[0].get_xlabel(), 'horizontal_x index')
        self.assertEqual(fig.figure.axes[0].get_ylabel(), 'Value')
        self.assertEqual(fig.figure.axes[0].get_xlim(), (0, 127))
        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[0].get_text(), 'Dataset 0')
        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[1].get_text(), 'Dataset 1')


    @patch('matplotlib.pyplot.show')
    def test_show1D_figure_subplot(self, mock_show):
        # Test all components of the plot

        try:
            fig = show1D([self.data, self.data2], slice_list=None)
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with all plot components: {e}")

        self.assertEqual(len(fig.figure.axes), 3)
        
        # Check the data
        line = fig.figure.axes[0].lines[0]
        np.testing.assert_array_equal(line.get_ydata(), self.data.array[:, 64, 64])

        line = fig.figure.axes[0].lines[1]
        np.testing.assert_array_equal(line.get_ydata(), self.data2.array[:, 64, 64])

        line = fig.figure.axes[1].lines[0]
        np.testing.assert_array_equal(line.get_ydata(), self.data.array[64, :, 64])

        line = fig.figure.axes[1].lines[1]
        np.testing.assert_array_equal(line.get_ydata(), self.data2.array[64, :, 64])

        line = fig.figure.axes[2].lines[0]
        np.testing.assert_array_equal(line.get_ydata(), self.data.array[64, 64, :])

        line = fig.figure.axes[2].lines[1]
        np.testing.assert_array_equal(line.get_ydata(), self.data2.array[64, 64, :])


        self.assertEqual(fig.figure.axes[0].get_title(), 'Slice at horizontal_y:64, horizontal_x:64')
        self.assertEqual(fig.figure.axes[1].get_title(), 'Slice at vertical:64, horizontal_x:64')
        self.assertEqual(fig.figure.axes[2].get_title(), 'Slice at vertical:64, horizontal_y:64')

        self.assertEqual(fig.figure.axes[0].get_xlabel(), 'vertical index')
        self.assertEqual(fig.figure.axes[1].get_xlabel(), 'horizontal_y index')
        self.assertEqual(fig.figure.axes[2].get_xlabel(), 'horizontal_x index')

        self.assertEqual(fig.figure.axes[0].get_ylabel(), 'Value')
        self.assertEqual(fig.figure.axes[1].get_ylabel(), 'Value')
        self.assertEqual(fig.figure.axes[2].get_ylabel(), 'Value')

        self.assertEqual(fig.figure.axes[0].get_xlim(), (0, 127))
        self.assertEqual(fig.figure.axes[1].get_xlim(), (0, 127))
        self.assertEqual(fig.figure.axes[2].get_xlim(), (0, 127))

        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[0].get_text(), 'Dataset 0')
        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[1].get_text(), 'Dataset 1')

        self.assertEqual(fig.figure.axes[1].get_legend().get_texts()[0].get_text(), 'Dataset 0')
        self.assertEqual(fig.figure.axes[1].get_legend().get_texts()[1].get_text(), 'Dataset 1')

        self.assertEqual(fig.figure.axes[2].get_legend().get_texts()[0].get_text(), 'Dataset 0')
        self.assertEqual(fig.figure.axes[2].get_legend().get_texts()[1].get_text(), 'Dataset 1')


    @patch('matplotlib.pyplot.show')
    def test_show1D_inputs_slice_list(self, mock_show):
        # Test with slice_list as wrong length
        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=[(1, 1)])
        self.assertEqual(str(cm.exception), "slice_list must provide a slice for ndim - 1 axes")

        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=[(0, 1), (1, 1), (2, 1)])
        self.assertEqual(str(cm.exception), "slice_list must provide a slice for ndim - 1 axes")

        # Test with slice_list with duplicate axis
        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=[(0, 1), ("Vertical", 2)])
        self.assertEqual(str(cm.exception), "slice_list contains duplicate axes. Each axis must be unique.")

        # Test with slice_list with invalid axis
        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=[("Vertical", 1),("No", 1)])
        self.assertEqual(str(cm.exception), f"Invalid axis label: No")

        try:
            show1D(self.data, slice_list=None)
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with single dataset and no slice_list: {e}")

        mock_show.reset_mock()

        try:
            show1D(self.data, slice_list=[("Horizontal_x", 30), ("Horizontal_y", 40)])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with single dataset and slice_list length 1: {e}")

        mock_show.reset_mock()

        try:
            show1D(self.data, slice_list=[(2, 30), (1, 40)])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with single dataset and slice_list length 1: {e}")

        mock_show.reset_mock()

        try:
            show1D(self.data, slice_list=[[(0, 62), (1, 70)], [(0, 61), (1, 70)], [(0, 67), (1, 70)]])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with single dataset and slice_list length 3: {e}")
   
        mock_show.reset_mock()

    @patch('matplotlib.pyplot.show')
    def test_show1D_inputs_datasets(self, mock_show):
        # Test with incompatible datasets
        with self.assertRaises(ValueError) as cm:
            show1D([self.data, self.data3])
        self.assertEqual(str(cm.exception), "All datasets must have the same shape")

        # Test with dataset_labels as wrong length
        with self.assertRaises(ValueError) as cm:
            show1D([self.data, self.data2], dataset_labels=["Data1"])
        self.assertEqual(str(cm.exception), "dataset_labels must be a list of strings equal to the number of datasets")

        with self.assertRaises(ValueError) as cm:
            show1D([self.data, self.data2], dataset_labels=["Data1", "Data2", "Data3"])
        self.assertEqual(str(cm.exception), "dataset_labels must be a list of strings equal to the number of datasets")


        try:
            fig = show1D([self.data, self.data2])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with default dataset_labels: {e}")

        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[0].get_text(), 'Dataset 0')
        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[1].get_text(), 'Dataset 1')
        mock_show.reset_mock()

        try:
            fig = show1D([self.data, self.data2], dataset_labels=["Data1", "Data2"])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with passed dataset_labels: {e}")

        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[0].get_text(), 'Data1')
        self.assertEqual(fig.figure.axes[0].get_legend().get_texts()[1].get_text(), 'Data2')
        mock_show.reset_mock()


    @patch('matplotlib.pyplot.show')
    def test_show1D_input_title(self, mock_show):

        # Test with title as wrong length
        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=None, title=["Title1"])
        self.assertEqual(str(cm.exception), "title must be a list of strings equal to the number of plots")

        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=[(0, 30), (1, 40)], title=["Title1", "Title2"])
        self.assertEqual(str(cm.exception), "title must be a list of strings equal to the number of plots")


        try:
            fig = show1D(self.data, slice_list=None)
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with default title: {e}")

        self.assertEqual(fig.figure.axes[0].get_title(), 'Slice at horizontal_y:64, horizontal_x:64')
        self.assertEqual(fig.figure.axes[1].get_title(), 'Slice at vertical:64, horizontal_x:64')
        self.assertEqual(fig.figure.axes[2].get_title(), 'Slice at vertical:64, horizontal_y:64')
        mock_show.reset_mock()

        try:
            fig = show1D(self.data, slice_list=None, title=["Title1", "Title2", "Title3"])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with passed title: {e}")

        self.assertEqual(fig.figure.axes[0].get_title(), 'Title1')
        self.assertEqual(fig.figure.axes[1].get_title(), 'Title2')
        self.assertEqual(fig.figure.axes[2].get_title(), 'Title3')


    @patch('matplotlib.pyplot.show')
    def test_show1D_input_axis_labels(self, mock_show):

        # Test with title as wrong length
        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=None, axis_labels=[("X", "Y")])
        self.assertEqual(str(cm.exception), "axis_labels must be a tuple or a list of tuples equal to the number of plots")

        with self.assertRaises(ValueError) as cm:
            show1D(self.data, slice_list=[(0, 30), (1, 40)], axis_labels=[("X", "Y"),("X", "Y"),("X", "Y")])
        self.assertEqual(str(cm.exception), "axis_labels must be a tuple or a list of tuples equal to the number of plots")

        try:
            fig = show1D(self.data, slice_list=None)
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with default axis_labels: {e}")

        self.assertEqual(fig.figure.axes[0].get_xlabel(), 'vertical index')
        self.assertEqual(fig.figure.axes[1].get_xlabel(), 'horizontal_y index')
        self.assertEqual(fig.figure.axes[2].get_xlabel(), 'horizontal_x index')

        self.assertEqual(fig.figure.axes[0].get_ylabel(), 'Value')
        self.assertEqual(fig.figure.axes[1].get_ylabel(), 'Value')
        self.assertEqual(fig.figure.axes[2].get_ylabel(), 'Value')

        mock_show.reset_mock()

        try:
            fig = show1D(self.data, slice_list=None, axis_labels=("X", "Y"))
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with passed axis_labels tuple: {e}")

        self.assertEqual(fig.figure.axes[0].get_xlabel(), 'X')
        self.assertEqual(fig.figure.axes[1].get_xlabel(), 'X')
        self.assertEqual(fig.figure.axes[2].get_xlabel(), 'X')
        self.assertEqual(fig.figure.axes[0].get_ylabel(), 'Y')
        self.assertEqual(fig.figure.axes[1].get_ylabel(), 'Y')
        self.assertEqual(fig.figure.axes[2].get_ylabel(), 'Y')

        mock_show.reset_mock()

        try:
            fig = show1D(self.data, slice_list=None, axis_labels=[("X1", "Y1"), ("X2", "Y2"), ("X3", "Y3")])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with passed axis_labels list: {e}")

        self.assertEqual(fig.figure.axes[0].get_xlabel(), 'X1')
        self.assertEqual(fig.figure.axes[1].get_xlabel(), 'X2')
        self.assertEqual(fig.figure.axes[2].get_xlabel(), 'X3')

        self.assertEqual(fig.figure.axes[0].get_ylabel(), 'Y1')
        self.assertEqual(fig.figure.axes[1].get_ylabel(), 'Y2')
        self.assertEqual(fig.figure.axes[2].get_ylabel(), 'Y3')


    @patch('matplotlib.pyplot.show')
    def test_show1D_inputs_line_colours(self, mock_show):
        
        import logging
        logging.basicConfig(level=logging.WARNING)
        log = logging.getLogger("cil.utilities.display")
        # Test with line_colours as wrong length
        with self.assertLogs(log, level='WARNING') as cm:
            fig = show1D([self.data, self.data2], line_colours=["red"])
            mock_show.assert_called_once()

        self.assertIn('line_colours must be a list of colours at least as long as the number of datasets, using default colour palette', cm.output[0])

        # defaults '#377eb8', '#ff7f00'
        self.assertEqual(fig.figure.axes[0].lines[0].get_color(), '#377eb8')
        self.assertEqual(fig.figure.axes[0].lines[1].get_color(), '#ff7f00')

        mock_show.reset_mock()

        try:
            fig = show1D([self.data, self.data2], line_colours=["red", "blue"])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with passed line_colours: {e}")

        self.assertEqual(fig.figure.axes[0].lines[0].get_color(), 'red')
        self.assertEqual(fig.figure.axes[0].lines[1].get_color(), 'blue')


    @patch('matplotlib.pyplot.show')
    def test_show1D_inputs_line_styles(self, mock_show):
        import logging
        logging.basicConfig(level=logging.WARNING)
        log = logging.getLogger("cil.utilities.display")
        # Test with line_styles wrong length
        with self.assertLogs(log, level='WARNING') as cm:
            fig = show1D([self.data, self.data2], line_styles=["--"])
        self.assertIn('line_styles must be a list of styles at least as long as the number of datasets, using default line styles', cm.output[0])

        # defaults '-', '-.'
        self.assertEqual(fig.figure.axes[0].lines[0].get_linestyle(), '-')
        self.assertEqual(fig.figure.axes[0].lines[1].get_linestyle(), '--')

        mock_show.reset_mock()
        try:
            fig = show1D([self.data, self.data2], line_styles=[":", "-."])
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with passed line_styles: {e}")

        self.assertEqual(fig.figure.axes[0].lines[0].get_linestyle(), ':')
        self.assertEqual(fig.figure.axes[0].lines[1].get_linestyle(), '-.')


    @patch('matplotlib.pyplot.show')
    def test_show1D_figure_size(self, mock_show):

        try:
            fig = show1D(self.data, slice_list=None)
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with default figsize: {e}")

        np.testing.assert_array_equal(fig.figure.get_size_inches(), (8,3*3))

        mock_show.reset_mock()
        try:
            fig = show1D(self.data, slice_list=None, size=(6, 6))
            mock_show.assert_called_once()
        except Exception as e:
            self.fail(f"show1D raised an exception with passed figsize: {e}")

        np.testing.assert_array_equal(fig.figure.get_size_inches(), (6,6*3))